#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function, annotations

from typing import Optional, List, Dict, Union, TYPE_CHECKING

if TYPE_CHECKING:
    import langchain
from dataclasses import dataclass
from datetime import datetime

import pandas

import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.wml_client_error import (WMLClientError, ValidationError,
                                                          InvalidValue, InvalidMultipleArguments, PromptVariablesError)
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.foundation_models.utils.enums import ModelTypes, PromptTemplateFormats
from ibm_watson_machine_learning.foundation_models.utils.utils import TemplateFormatter


@dataclass
class PromptTemplateLock:
    """Storage for lock object.
    """
    locked: bool
    locked_by: Optional[str] = None


class PromptTemplate:
    """Storage for prompt template parameters.

    :param prompt_id: Id of prompt template, defaults to None.
    :type prompt_id: Optional[str], attribute setting not allowed

    :param created_at: Time the prompt was created (UTC), defaults to None.
    :type created_at: Optional[str], attribute setting not allowed 

    :param lock: Locked state of asset, defaults to None.
    :type lock: Optional[PromptTemplateLock], attribute setting not allowed

    :param is_template: True if prompt is a template, False otherwise; defaults to None.
    :type is_template: Optional[bool], attribute setting not allowed 

    :param name: Prompt template name, defaults to None.
    :type name: Optional[str], optional

    :param model_id: Foundation model id, defaults to None.
    :type model_id: Optional[ModelTypes], optional

    :param model_params: Model parameters, defaults to None.
    :type model_params: Optional[Dict], optional

    :param template_version: Semvar version for tracking in IBM AI Factsheets, defaults to None.
    :type template_version: Optional[str], optional

    :param task_ids: List of task ids, defaults to None.
    :type task_ids: Optional[List[str]], optional

    :param description: Prompt template asset description, defaults to None.
    :type description: Optional[str], optional

    :param input_text: Input text for prompt, defaults to None.
    :type input_text: Optional[str], optional

    :param input_variables: Input variables can be present in fields: `instruction`, 
                            `input_prefix`, `output_prefix`, `input_text`, `examples`
                            and are indentified by braces ('{' and '}'), defaults to None.
    :type temaplate_parameters: (List | Dict[str, Dict[str, str]] | None), optional

    :param instruction: Instruction for model, defaults to None.
    :type instruction: Optional[str], optional

    :param input_prefix: Prefix string placed before input text, defaults to None.
    :type input_prefix: Optional[str], optional

    :param output_prefix: Prefix before model response, defaults to None.
    :type output_prefix: Optional[str], optional

    :param exmaples: Examples may help the model to adjust the response; [[input1, output1], ...], defaults to None.
    :type exmaples: Optional[List[List[str]]], optional

    :param validate_template: If True, the Prompt Template is validated for the presence of input variables, defaults to True.
    :type validate_template: bool, optional


    **Examples**

    Example of invalid Prompt Template:

    .. code-block:: python

        prompt_template = PromptTemplate(input_text='What are the most famous monuments in ?',
                                         input_variables=['country'])

        Traceback (most recent call last):
            ...
        ValidationError: Invalid prompt template; check for mismatched or missing input variables. Missing input variable: {'country'}

    Example of valid Prompt Template:

    .. code-block:: python

        prompt_template = PromptTemplate(input_text='What are the most famous monuments in {country}?',
                                         input_variables=['country'])

    """

    def __init__(self,
                 name: Optional[str] = None,
                 model_id: Optional[ModelTypes] = None,
                 model_params: Optional[Dict] = None,
                 template_version: Optional[str] = None,
                 task_ids: Optional[List[str]] = None,
                 description: Optional[str] = None,
                 input_text: Optional[str] = None,
                 input_variables: (List | Dict[str, Dict[str, str]] | None) = None,
                 instruction: Optional[str] = None,
                 input_prefix: Optional[str] = None,
                 output_prefix: Optional[str] = None,
                 examples: Optional[List[List[str]]] = None,
                 validate_template: bool = True) -> None:
        self.name = name
        self._prompt_id = None
        self._created_at = None
        self._lock = None
        self._is_template = None
        self.model_id = model_id
        if isinstance(self.model_id, ModelTypes):
            self.model_id = self.model_id.value
        self.model_params = model_params.copy() if model_params is not None else model_params
        self.task_ids = task_ids.copy() if task_ids is not None else task_ids
        self.template_version = template_version
        self.description = description
        self.input_text = input_text
        self.input_variables = input_variables.copy() if input_variables is not None else input_variables
        self.instruction = instruction
        self.input_prefix = input_prefix
        self.output_prefix = output_prefix
        self.examples = examples.copy() if examples is not None else examples

        # template validation
        if validate_template:
            self._validation()

    def __repr__(self):
        args = [f"{key}={value!r}" for key, value in self.__dict__.items() 
                if not key.startswith('_') and value is not None]
        return f"{type(self).__name__}({ ', '.join(args)})"
    
    @property
    def prompt_id(self):
        return self._prompt_id

    @property
    def created_at(self):
        return str(datetime.utcfromtimestamp(self._created_at / 1000)).split(".")[0]

    @property
    def lock(self):
        return self._lock

    @property
    def is_template(self):
        return self._is_template

    def _validation(self):
        """Validate template structure.

        :raises ValidationError: raises when input_variables does not fit placeholders in input body.
        """
        input_variables = self.input_variables if self.input_variables is not None else []
        template_text = " ".join(filter(None, [self.instruction,
                                               self.input_prefix,
                                               self.output_prefix]))
        if self.examples:
            for example in self.examples:
                template_text += " ".join(example)
        try:
            def _validate(input_text):
                dummy_inputs = {input_variable: "wx" for input_variable in input_variables}
                TemplateFormatter().format("".join([template_text, input_text]), **dummy_inputs)

            if self.input_text:
                _validate(template_text + self.input_text)
            else:
                if template_text:
                    _validate(template_text)
        except KeyError as key:
            raise ValidationError(key)


class PromptTemplateManager(WMLResource):
    """Instantiate the prompt template manager.

    :param credentials: Credentials to watsonx.ai instance.
    :type credentials: dict

    :param project_id: ID of project
    :type project_id: str

    :param space_id: ID of project
    :type space_id: str

    :param verify: user can pass as verify one of following:
        - the path to a CA_BUNDLE file
        - the path of directory with certificates of trusted CAs
        - `True` - default path to truststore will be taken
        - `False` - no verification will be made
    :type verify: bool or str, optional

    .. note::
        One of these parameters is required: ['project_id ', 'space_id']

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models.prompts import PromptTemplate, PromptTemplateManager
        from ibm_watson_machine_learning.foundation_models.utils.enums import ModelTypes

        prompt_mgr = PromptTemplateManager(
                        credentials={
                            "apikey": "***",
                            "url": "https://us-south.ml.cloud.ibm.com"
                        },
                        project_id="*****"
                        )

        prompt_template = PromptTemplate(name="My prompt",
                                         model_id=ModelTypes.GRANITE_13B_CHAT_V2,
                                         input_prefix="Human:",
                                         output_prefix="Assistant:",
                                         input_text="What is {object} and how does it work?",
                                         input_variables=['object'],
                                         examples=[['What is the Stock Market?',
                                                    'A stock market is a place where investors buy and sell shares of publicly traded companies.']])

        stored_prompt_template = prompt_mgr.store_prompt(prompt_template)
        print(stored_prompt_template.prompt_id)   # id of prompt template asset
    """

    def __init__(self,
                 credentials: Optional[dict] = None,
                 *,
                 project_id: Optional[str] = None,
                 space_id: Optional[str] = None,
                 verify=None,
                 api_client: APIClient = None) -> None:

        self.project_id = project_id
        self.space_id = space_id
        if credentials:
            self._client = APIClient(credentials, verify=verify)
        elif api_client:
            self._client = api_client
        else:
            raise InvalidMultipleArguments(params_names_list=["credentials", "api_client"],
                                           reason="None of the arguments were provided.")

        if self.space_id is not None and self.project_id is not None:
            raise InvalidMultipleArguments(params_names_list=["project_id", "space_id"],
                                           reason="Both arguments were provided.")
        self.params = {}
        if self.space_id:
            self._client.set.default_space(space_id)
            self.params = {'space_id': self.space_id}
        elif self.project_id:
            self._client.set.default_project(project_id)
            self.params = {'project_id': self.project_id}
        elif api_client:
            if (project_id := self._client.default_project_id):
                self.params = {'project_id': project_id}
            elif (space_id := self._client.default_space_id):
                self.params = {'space_id': space_id}
            else:
                pass
        elif not api_client:
            raise InvalidMultipleArguments(params_names_list=["space_id", "project_id"],
                                           reason="None of the arguments were provided.")
        
        if not self._client.CLOUD_PLATFORM_SPACES and self._client.CPD_version < 4.8:
            raise WMLClientError(error_msg="Operation is unsupported for this release.")

        WMLResource.__init__(self, __name__, self._client)

    def _create_request_body(self, prompt_template: PromptTemplate) -> Dict:
        """Method is used to create request body from PromptTemplate object.

        :param prompt_template: Object of type PromptTemplate based on which the request
                                body will be created.
        :type prompt_template: PromptTemplate

        :return: Request body
        :rtype: Dict
        """
        json_data: Dict = {'prompt': dict()}
        if prompt_template.description is not None:
            json_data.update({'description': prompt_template.description})
        if prompt_template.input_variables is not None:
            PromptTemplateManager._validate_type(prompt_template.input_variables, u'input_variables', [dict, list],
                                                 False)
            if isinstance(prompt_template.input_variables, list):
                json_data.update({'prompt_variables': {key: {} for key in prompt_template.input_variables}})
            else:
                json_data.update({'prompt_variables': prompt_template.input_variables})
        if prompt_template.task_ids is not None:
            PromptTemplateManager._validate_type(prompt_template.task_ids, u'task_ids', list, False)
            json_data.update({'task_ids': prompt_template.task_ids})
        if prompt_template.template_version is not None:
            json_data.update({"model_version": {"number": prompt_template.template_version}})

        if prompt_template.input_text:
            PromptTemplateManager._validate_type(prompt_template.input_text, u'input_text', str, False)
            json_data['prompt'].update({'input': [[prompt_template.input_text, '']]})

        PromptTemplateManager._validate_type(prompt_template.model_id, u'model_id', str, True)
        if prompt_template.model_id is not None:
            json_data['prompt'].update({'model_id': prompt_template.model_id})

        if prompt_template.model_params is not None:
            PromptTemplateManager._validate_type(prompt_template.model_params, u'model_parameters', dict, False)
            json_data['prompt'].update({'model_parameters': prompt_template.model_params})

        data: Dict = dict()
        if prompt_template.instruction is not None:
            data.update({'instruction': prompt_template.instruction})

        if prompt_template.input_prefix is not None:
            data.update({'input_prefix': prompt_template.input_prefix})

        if prompt_template.output_prefix is not None:
            data.update({'output_prefix': prompt_template.output_prefix})
        if prompt_template.examples is not None:
            PromptTemplateManager._validate_type(prompt_template.examples, u'examples', list, False)
            data.update({'examples': prompt_template.examples})

        json_data['prompt'].update({'data': data})

        return json_data

    def _from_json_to_prompt(self, response: Dict) -> PromptTemplate:
        """Convert json response to PromptTemplate object.

        :param response: Response body after request operation.
        :type response: Dict

        :return: PromptTemplate object with given details.
        :rtype: PromptTemplate
        """
        prompt_field: Dict = response.get('prompt', dict())
        data_field: Dict = prompt_field.get('data', dict())
        prompt_template = PromptTemplate(name=response.get('name'),
                                         description=response.get('description'),
                                         model_id=prompt_field.get('model_id'),
                                         model_params=prompt_field.get('model_parameters'),
                                         task_ids=response.get("task_ids"),
                                         template_version=response.get("model_version", dict()).get("number"),
                                         input_variables = response.get('prompt_variables'),
                                         input_text = prompt_field.get('input', [[None, None]])[0][0],
                                         instruction = data_field.get('instruction'),
                                         input_prefix = data_field.get('input_prefix'),
                                         output_prefix = data_field.get('output_prefix'),
                                         examples = data_field.get('examples'),
                                         validate_template=False
                                         )

        prompt_template._prompt_id = response.get('id')
        prompt_template._created_at = response.get('created_at')
        if "lock_type" in response.get('lock', dict()):
            del response['lock']["lock_type"]
        prompt_template._lock = PromptTemplateLock(**response.get('lock', {"locked": None,
                                                                           "locked_by": None}))
        prompt_template._is_template = response.get('is_template')

        return prompt_template

    def _get_details(self, limit: Optional[int] = None) -> List:
        """Method retrives details of all prompt templates. If limit is set to None
        then all prompt templates are fetched.

        :param limit: limit number of fetched records, defaults to None.
        :type limit: Optional[int]

        :return: List of prompts metadata 
        :rtype: List
        """
        headers = self._client._get_headers()
        url = self._client.service_instance._href_definitions.get_prompts_all_href()
        json_data = {"query": u"asset.asset_type:wx_prompt",
                     u"sort": "-asset.created_at<string>"}
        if limit is not None:
            if limit < 1:
                raise WMLClientError('Limit cannot be lower than 1.')
            elif limit > 200:
                raise WMLClientError('Limit cannot be larger than 200.')

            json_data.update({'limit': limit})
        else:
            json_data.update({'limit': 200})
        prompts_list = []
        bookmark = True
        while bookmark is not None:
            response = requests.post(url=url, json=json_data,
                                     headers=headers,
                                     params=self.params)
            details_json = self._handle_response(200, "Get next details", response)
            bookmark = details_json.get('next', {'href': None}).get('bookmark', None)
            prompts_list.extend(details_json.get('results', []))
            if limit is not None:
                break
            json_data.update({'bookmark': bookmark})
        return prompts_list

    def _change_lock(self, prompt_id: str, locked: bool, force: bool = False) -> Dict:
        """Change prompt template lock state.

        :param prompt_id: Id of prompt template.
        :type prompt_id: str

        :param locked: New lock state.
        :type locked: bool

        :param force: force lock state overwrite, defaults to False.
        :type force: bool, optional

        :return: Response content after lock state change.
        :rtype: Dict
        """
        headers = self._client._get_headers()
        params = (self.params | {"prompt_id": prompt_id, "force": force})
        json_data = {"locked": locked}

        url = self._client.service_instance._href_definitions.get_prompts_href() + f"/{prompt_id}/lock"
        response = requests.put(url=url,
                                json=json_data,
                                headers=headers,
                                params=params)

        return self._handle_response(200, u'change_lock', response)

    def load_prompt(self,
                    prompt_id: str,
                    astype: PromptTemplateFormats = PromptTemplateFormats.PROMPTTEMPLATE,
                    *,
                    prompt_variables: Optional[Dict[str, str]] = None):
        """Retrive a prompt template asset.

        :param prompt_id: Id of prompt template which is processed.
        :type prompt_id: str

        :param astype: Type of return object.
        :type astype: PromptTemplateFormats

        :param prompt_variables: Dictionary of input variables and values with which input variables will be replaced.
        :type prompt_variables: Dict[str, str]

        :return: Prompt template asset.
        :rtype: PromptTemplate | str | langchain.prompts.PromptTemplate

        **Example**

        .. code-block:: python

            loaded_prompt_template = prompt_mgr.load_prompt(prompt_id)
            loaded_prompt_template_lc = prompt_mgr.load_prompt(prompt_id, PromptTemplateFormats.LANGCHAIN)
            loaded_prompt_template_string = prompt_mgr.load_prompt(prompt_id, PromptTemplateFormats.STRING)
        """
        headers = self._client._get_headers()
        params = (self.params | {"prompt_id": prompt_id})
        url = self._client.service_instance._href_definitions.get_prompts_href() + f"/{prompt_id}"

        if isinstance(astype, PromptTemplateFormats):
            astype = astype.value

        if astype == 'prompt':
            response = requests.get(url=url,
                                    headers=headers,
                                    params=params)
            return self._from_json_to_prompt(self._handle_response(200, u'_load_json_prompt', response))
        elif astype in ('langchain', 'string'):
            response = requests.post(url=url + u"/input",
                                     headers=headers,
                                     params=params)
            response_input = self._handle_response(200, u'load_prompt', response).get("input")
            if astype == 'string':
                try:
                    return response_input if prompt_variables is None else response_input.format(**prompt_variables)
                except KeyError as key:
                    raise PromptVariablesError(key)
            else:
                from langchain.prompts import PromptTemplate as LcPromptTemplate
                return LcPromptTemplate.from_template(response_input)
        else:
            raise InvalidValue(u'astype')

    def list(self, *, limit=None) -> pandas.core.frame.DataFrame:
        """List all available prompt templates in the DataFrame format.

        :param limit: limit number of fetched records, defaults to None.
        :type limit: Optional[int]

        :return: Dataframe of fundamental properties of availabale prompts.
        :rtype: pandas.core.frame.DataFram

        **Example**

        .. code-block:: python

            prompt_mgr.list(limit=5)    # list of 5 recent created prompt template assets

        .. hint::
            Additionally you can sort available prompt templates by "LAST MODIFIED" field.

            .. code-block:: python

                df_prompts = prompt_mgr.list()
                df_prompts.sort_values("LAST MODIFIED", ascending=False)

        """
        details = ['metadata.asset_id', 'metadata.name', 'metadata.created_at', 'metadata.usage.last_updated_at']
        prompts_details = self._get_details(limit=limit)

        data_normalize = pandas.json_normalize(prompts_details)
        prompts_data = data_normalize.reindex(columns=details)

        df_details = pandas.DataFrame(prompts_data, columns=details)

        df_details.rename(columns={'metadata.asset_id': 'ID',
                                   'metadata.name': 'NAME',
                                   'metadata.created_at': 'CREATED',
                                   'metadata.usage.last_updated_at': 'LAST MODIFIED',
                                   }, inplace=True)

        return df_details

    def store_prompt(self, prompt_template: Union[PromptTemplate, langchain.prompts.PromptTemplate]) -> PromptTemplate:
        """Store a new prompt template.

        :param prompt_template: PromptTemplate to be stored.
        :type prompt_template: (PromptTemplate | langchain.prompts.PromptTemplate)

        :return: PromptTemplate object initialized with values provided in the server response object.
        :rtype: PromptTemplate
        """
        if isinstance(prompt_template, PromptTemplate):
            pass
        else:
            from langchain.prompts import PromptTemplate as LcPromptTemplate
            if isinstance(prompt_template, LcPromptTemplate):
                prompt_template = PromptTemplate(name="My prompt",
                                                 model_id=ModelTypes.FLAN_UL2,
                                                 input_text=prompt_template.template,
                                                 input_variables=prompt_template.input_variables)
            else:
                raise WMLClientError(error_msg="Unsupported type for `prompt_template`")
            
        headers = self._client._get_headers()

        PromptTemplateManager._validate_type(prompt_template.name,
                                             u'prompt_template.name',
                                             str, True)
        json_data: Dict = {
            "name": prompt_template.name,
            "lock": {"locked": True},
            "prompt": dict()
        }

        json_data.update(self._create_request_body(prompt_template))

        url = self._client.service_instance._href_definitions.get_prompts_href()
        response = requests.post(url=url,
                                 json=json_data,
                                 headers=headers,
                                 params=self.params)
        response = self._handle_response(201, u'store_prompt', response)

        return self._from_json_to_prompt(response)

    def delete_prompt(self, prompt_id: str, *, force: bool = False) -> str:
        """Remove prompt template from project or space.

        :param prompt_id: Id of prompt template that will be delete. 
        :type prompt_id: str

        :param force: If True then prompt template is unlocked and then delete, defaults to False.
        :type force: bool

        :return: Status 'SUCESS' if the prompt template is successfully deleted.
        :rtype: str
        
        **Example**

        .. code-block:: python

            prompt_mgr.delete_prompt(prompt_id)  # delete if asset is unclocked
        """
        if force:
            self.unlock(prompt_id)

        headers = self._client._get_headers()
        params = (self.params | {"prompt_id": prompt_id})

        url = self._client.service_instance._href_definitions.get_prompts_href() + f"/{prompt_id}"
        response = requests.delete(url=url, headers=headers,
                                   params=params)

        return self._handle_response(204, u'delete_prompt', response)

    def update_prompt(self, prompt_id: str, prompt_template: PromptTemplate) -> Dict:
        """Update prompt template data.

        :param prompt_id: Id of the updated prompt template.
        :type prompt_id: str

        :param prompt: PromptTemplate with new data.
        :type prompt: PromptTemplate 

        :return: metadata of updated deployment
        :rtype: dict

        **Example**

        .. code-block:: python

            updataed_prompt_template = PromptTemplate(name="New name")
            prompt_mgr.update_prompt(prompt_id, prompt_template)  # {'name': 'New name'} in metadata  
            
        """
        headers = self._client._get_headers()
        params = (self.params | {"prompt_id": prompt_id})

        new_body: Dict = dict()
        current_prompt_template = self.load_prompt(prompt_id)

        for attribute in prompt_template.__dict__:
            if getattr(prompt_template, attribute) is not None and not attribute.startswith("_"):
                setattr(current_prompt_template, attribute, getattr(prompt_template, attribute))

        if current_prompt_template.name is not None:
            new_body.update({'name': current_prompt_template.name})

        new_body.update(self._create_request_body(current_prompt_template))

        url = self._client.service_instance._href_definitions.get_prompts_href() + f"/{prompt_id}"

        response = requests.patch(url=url,
                                  json=new_body,
                                  headers=headers,
                                  params=params)
        return self._handle_response(200, u'update_prompt', response)

    def get_lock(self, prompt_id: str) -> Dict:
        """Get the current locked state of a prompt template.

        :param prompt_id: Id of prompt template
        :type prompt_id: str

        :return: Information about locked state of prompt template asset.
        :rtype: Dict

        **Example**

        .. code-block:: python

            print(prompt_mgr.get_lock(prompt_id))
        """
        headers = self._client._get_headers()
        params = (self.params | {"prompt_id": prompt_id})
        url = self._client.service_instance._href_definitions.get_prompts_href() + f"/{prompt_id}/lock"

        response = requests.get(url=url,
                                headers=headers,
                                params=params)

        return self._handle_response(200, u'get_lock', response)

    def lock(self, prompt_id: str, force: bool = False) -> Dict:
        """Lock the prompt template if it is unlocked and user has permission to do that. 

        :param promp_id: Id of prompt template.
        :type promp_id: str

        :param force: If True, method forcefully overwrite a lock.
        :type force: bool

        :return: Status 'SUCCESS' or response content after an attempt to lock prompt template.
        :rtype: (str | Dict)

        **Example**

        .. code-block:: python

            prompt_mgr.lock(prompt_id)

        """
        return self._change_lock(prompt_id=prompt_id,
                                 locked=True,
                                 force=force)

    def unlock(self, prompt_id: str) -> Dict:
        """Unlock the prompt template if it is locked and user has permission to do that.

        :param promp_id: Id of prompt template.
        :type promp_id: str

        :return: Response content after an attempt to unlock prompt template.
        :rtype: Dict

        **Example**

        .. code-block:: python

            prompt_mgr.unlock(prompt_id)
        """
        # server returns status code 400 after trying to unlock unlocked prompt
        lock_state = self.get_lock(prompt_id)
        if lock_state['locked']:
            return self._change_lock(prompt_id=prompt_id,
                                     locked=False,
                                     force=False)
        else:
            return lock_state



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from enum import Enum

__all__ = [
    "ModelTypes",
    "DecodingMethods",
    "PromptTuningTypes",
    "PromptTuningInitMethods",
    "TuneExperimentTasks",
    "PromptTemplateFormats",
]


class ModelTypes(Enum):
    """Supported foundation models."""
    FLAN_T5_XXL = "google/flan-t5-xxl"
    FLAN_UL2 = "google/flan-ul2"
    MT0_XXL = "bigscience/mt0-xxl"
    GPT_NEOX = 'eleutherai/gpt-neox-20b'
    MPT_7B_INSTRUCT2 = 'ibm/mpt-7b-instruct2'
    STARCODER = 'bigcode/starcoder'
    LLAMA_2_70B_CHAT = 'meta-llama/llama-2-70b-chat'
    LLAMA_2_13B_CHAT = 'meta-llama/llama-2-13b-chat'
    GRANITE_13B_INSTRUCT = 'ibm/granite-13b-instruct-v1'
    GRANITE_13B_CHAT = 'ibm/granite-13b-chat-v1'
    FLAN_T5_XL = 'google/flan-t5-xl'
    GRANITE_13B_CHAT_V2 = 'ibm/granite-13b-chat-v2'
    GRANITE_13B_INSTRUCT_V2 = 'ibm/granite-13b-instruct-v2'
    ELYZA_JAPANESE_LLAMA_2_7B_INSTRUCT = 'elyza/elyza-japanese-llama-2-7b-instruct'
    MIXTRAL_8X7B_INSTRUCT_V01_Q = 'ibm-mistralai/mixtral-8x7b-instruct-v01-q'

class DecodingMethods(Enum):
    """Supported decoding methods for text generation."""
    SAMPLE = "sample"
    GREEDY = "greedy"


class PromptTuningTypes:
    PT = "prompt_tuning"


class PromptTuningInitMethods:
    """Supported methods for prompt initialization in prompt tuning."""
    RANDOM = "random"
    TEXT = "text"
    # PRESET ?


class TuneExperimentTasks(Enum):
    QUESTION_ANSWERING = "question_answering"
    SUMMARIZATION = "summarization"
    RETRIEVAL_AUGMENTED_GENERATION = "retrieval_augmented_generation"
    CLASSIFICATION = "classification"
    GENERATION = "generation"
    CODE_GENERATION_AND_CONVERSION = "code"
    EXTRACTION = "extraction"


class PromptTemplateFormats(Enum):
    """Supported formats of loaded prompt template."""
    PROMPTTEMPLATE = "prompt"
    STRING = "string"
    LANGCHAIN = "langchain"



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from .model import Model
from .prompt_tuner import PromptTuner
from ibm_watson_machine_learning.foundation_models.utils.utils import (get_model_specs, get_model_lifecycle,
                                                                       get_model_specs_with_prompt_tuning_support,
                                                                       get_supported_tasks, get_custom_model_specs)
from ibm_watson_machine_learning.foundation_models.inference.model_inference import ModelInference



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems


class FrameworkOutputRepository(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None, version=None ,runtime= None, runtimes = None, libraries = None):
        """
        FrameworkOutputRepository - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'version': 'str',
            'runtime' : 'FrameworkOutputRepositoryRuntimes',
            'runtimes': 'list[FrameworkOutputRepositoryRuntimes]',
            'libraries': 'list[FrameworkOutputRepositoryLibraries]'
        }

        self.attribute_map = {
            'name': 'name',
            'version': 'version',
            'runtime' : 'runtime',
            'runtimes': 'runtimes',
            'libraries': 'libraries'
        }

        self._name = name
        self._version = version
        self._runtime = runtime
        self._runtimes = runtimes
        self._libraries = libraries

    @property
    def name(self):
        """
        Gets the name of this FrameworkOutputRepository.


        :return: The name of this FrameworkOutputRepository.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this FrameworkOutputRepository.


        :param name: The name of this FrameworkOutputRepository.
        :type: str
        """
        self._name = name

    @property
    def version(self):
        """
        Gets the version of this FrameworkOutputRepository.


        :return: The version of this FrameworkOutputRepository.
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """
        Sets the version of this FrameworkOutputRepository.


        :param version: The version of this FrameworkOutputRepository.
        :type: str
        """
        self._version = version

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result


    @property
    def runtimes(self):
        """
        Gets the runtimes of this FrameworkOutputRepository.


        :return: The runtimes of this FrameworkOutputRepository.
        :rtype: list[FrameworkOutputRepositoryRuntimes]
        """
        return self._runtimes

    @runtimes.setter
    def runtimes(self, runtimes):
        """
        Sets the runtimes of this FrameworkOutputRepository.


        :param runtimes: The runtimes of this FrameworkOutputRepository.
        :type: list[FrameworkOutputRepositoryRuntimes]
        """
        self._runtimes = runtimes

    @property
    def libraries(self):
        """
        Gets the libraries of this FrameworkOutputRepository.


        :return: The libraries of this FrameworkOutputRepository.
        :rtype: list[FrameworkOutputRepositoryLibraries]
        """

        return self._libraries

    @libraries.setter
    def libraries(self, libraries):
        """
        Sets the libraries of this FrameworkOutputRepository.

        :param runtimes: The runtimes of this FrameworkOutputRepository.
        :type: list[FrameworkOutputRepositoryLibraries]
        """
        self._libraries = libraries


    @property
    def runtime(self):
        """
        Gets the runtime of this FrameworkOutputRepository.


        :return: The runtimes of this FrameworkOutputRepository.
        :rtype: list[FrameworkOutputRepositoryRuntimes]
        """
        return self._runtime

    @runtime.setter
    def runtime(self, runtime):
        """
        Sets the runtimes of this FrameworkOutputRepository.


        :param runtimes: The runtimes of this FrameworkOutputRepository.
        :type: list[FrameworkOutputRepositoryRuntimes]
        """
        self._runtimes = runtime

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import re
import os
import sys
import shutil
import tarfile
import logging
import importlib.util
import json
import numpy
import inspect
import importlib

from typing import Optional
from subprocess import check_call
from packaging import version

import ibm_watson_machine_learning._wrappers.requests as requests

from ibm_watson_machine_learning.wml_client_error import WMLClientError, CannotInstallLibrary

INSTANCE_DETAILS_TYPE = u'instance_details_type'
PIPELINE_DETAILS_TYPE = u'pipeline_details_type'
DEPLOYMENT_DETAILS_TYPE = u'deployment_details_type'
EXPERIMENT_RUN_DETAILS_TYPE = u'experiment_run_details_type'
MODEL_DETAILS_TYPE = u'model_details_type'
DEFINITION_DETAILS_TYPE = u'definition_details_type'
EXPERIMENT_DETAILS_TYPE = u'experiment_details_type'
TRAINING_RUN_DETAILS_TYPE = u'training_run_details_type'
FUNCTION_DETAILS_TYPE = u'function_details_type'
DATA_ASSETS_DETAILS_TYPE = u'data_assets_details_type'
SW_SPEC_DETAILS_TYPE = u'sw_spec_details_type'
HW_SPEC_DETAILS_TYPE = u'hw_spec_details_type'
RUNTIME_SPEC_DETAILS_TYPE = u'runtime_spec_details_type'
LIBRARY_DETAILS_TYPE = u'library_details_type'
SPACES_DETAILS_TYPE = u'spaces_details_type'
MEMBER_DETAILS_TYPE = u'member_details_type'
CONNECTION_DETAILS_TYPE = u'connection_details_type'
PKG_EXTN_DETAILS_TYPE = u'pkg_extn_details_type'
UNKNOWN_ARRAY_TYPE = u'resource_type'
UNKNOWN_TYPE = u'unknown_type'
SPACES_IMPORTS_DETAILS_TYPE = u'spaces_imports_details_type'
SPACES_EXPORTS_DETAILS_TYPE = u'spaces_exports_details_type'

SPARK_MLLIB = u'mllib'
SPSS_FRAMEWORK = u'spss-modeler'
TENSORFLOW_FRAMEWORK = u'tensorflow'
XGBOOST_FRAMEWORK = u'xgboost'
SCIKIT_LEARN_FRAMEWORK = u'scikit-learn'
PMML_FRAMEWORK = u'pmml'


def is_python_2():
    return sys.version_info[0] == 2


def get_url(url, headers, params=None, isIcp=False):
    import ibm_watson_machine_learning._wrappers.requests as requests

    if isIcp:
        return requests.get(url, headers=headers, params=params)
    else:
        return requests.get(url, headers=headers, params=params)


def print_text_header_h1(title):
    print(u'\n\n' + (u'#' * len(title)) + u'\n')
    print(title)
    print(u'\n' + (u'#' * len(title)) + u'\n\n')


def print_text_header_h2(title):
    print(u'\n\n' + (u'-' * len(title)))
    print(title)
    print((u'-' * len(title)) + u'\n\n')


def get_type_of_details(details):
    if 'resources' in details:
        return UNKNOWN_ARRAY_TYPE
    elif details is None:
        raise WMLClientError('Details doesn\'t exist.')
    else:
        try:
            plan = 'plan' in details[u'entity']

            if plan:
                return INSTANCE_DETAILS_TYPE

            if re.search(u'\/wml_instances\/[^\/]+$', details[u'metadata'][u'url']) is not None:
                return INSTANCE_DETAILS_TYPE
        except:
            pass
        try:
            if re.search(u'\/pipelines\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return PIPELINE_DETAILS_TYPE
        except:
            pass
        try:
            if 'href' in details[u'metadata'] and re.search(u'\/deployments\/[^\/]+$', details[u'metadata'][u'href']) is not None \
                    or re.search(u'\/deployments\/[^\/]+$', details[u'metadata'][u'id']) is not None \
                    or u'virtual_deployment_downloads' in details[u'entity'][u'status']:
                return DEPLOYMENT_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/experiments\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return EXPERIMENT_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/trainings\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return TRAINING_RUN_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/models\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return MODEL_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/functions\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return FUNCTION_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/runtimes\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return RUNTIME_SPEC_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/libraries\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return LIBRARY_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/spaces\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return SPACES_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/members\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return MEMBER_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/members\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return MEMBER_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/assets\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return DATA_ASSETS_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/software_specifications\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return SW_SPEC_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/hardware_specifications\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return HW_SPEC_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/package_extension\/[^\/]+$', details[u'entity'][u'package_extension'][u'href']) is not None:
                return PKG_EXTN_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/imports\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return SPACES_IMPORTS_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/exports\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return SPACES_EXPORTS_DETAILS_TYPE
        except:
            pass

        return UNKNOWN_TYPE


def load_model_from_directory(framework, directory_path):
    if "mllib" in framework:
     from pyspark.ml import PipelineModel
     return PipelineModel.read().load(directory_path)
    if "spss" in framework:
     pass
    if "tensorflow" in framework:
     pass
    if "scikit" in framework or "xgboost" in framework:
        try:
            try:
                from sklearn.externals import joblib
            except ImportError:
                import joblib
            pkl_files = [x for x in os.listdir(directory_path) if x.endswith('.pkl')]

            if len(pkl_files) < 1:
                raise WMLClientError('No pkl files in directory.')

            model_id = pkl_files[0]
            return joblib.load(os.path.join(directory_path, model_id))
        except Exception as e:
            raise WMLClientError('Cannot load model from pkl file.', e)
    if "pmml" in framework:
     pass
    else:
        raise WMLClientError(u'Invalid framework specified: \'{}\'.'.format(framework))


# def load_model_from_directory(framework, directory_path):
#     if framework == SPARK_MLLIB:
#         from pyspark.ml import PipelineModel
#         return PipelineModel.read().load(directory_path)
#     elif framework == SPSS_FRAMEWORK:
#         pass
#     elif framework == TENSORFLOW_FRAMEWORK:
#         pass
#     elif framework == SCIKIT_LEARN_FRAMEWORK or framework == XGBOOST_FRAMEWORK:
#         from sklearn.externals import joblib
#         model_id = directory_path[directory_path.rfind('/') + 1:] + ".pkl"
#         return joblib.load(os.path.join(directory_path, model_id))
#     elif framework == PMML_MODEL:
#         pass
#     else:
#         raise WMLClientError('Invalid framework specified: \'{}\'.'.format(framework))


def save_model_to_file(model, framework, base_path, filename):
    if filename.find('.') != -1:
        base_name = filename[:filename.find('.') + 1]
        file_extension = filename[filename.find('.'):]
    else:
        base_name = filename
        file_extension = 'tar.gz'

    if framework == SPARK_MLLIB:
        model.write.overwrite.save(os.path.join(base_path, base_name))
    elif framework == SPSS_FRAMEWORK:
        pass
    elif framework == TENSORFLOW_FRAMEWORK:
        pass
    elif framework == XGBOOST_FRAMEWORK:
        pass
    elif framework == SCIKIT_LEARN_FRAMEWORK:
        os.makedirs(os.path.join(base_path, base_name))
        try:
            from sklearn.externals import joblib
        except ImportError:
            import joblib
        joblib.dump(model, os.path.join(base_path, base_name, base_name + ".pkl"))
    elif framework == PMML_FRAMEWORK:
        pass
    else:
        raise WMLClientError(u'Invalid framework specified: \'{}\'.'.format(framework))


def format_metrics(latest_metrics_list):
    formatted_metrics = u''

    for i in latest_metrics_list:

        values = i[u'values']

        if len(values) > 0:
            sorted_values = sorted(values, key=lambda k: k[u'name'])
        else:
            sorted_values = values

        for j in sorted_values:
            formatted_metrics = formatted_metrics + i[u'phase'] + ':' + j[u'name']+'='+'{0:.4f}'.format(j[u'value']) + '\n'

    return formatted_metrics


def inherited_docstring(f, mapping={}):
    def dec(obj):
        if not obj.__doc__:
            possible_types = {'model': 'model',
                              'function': 'function',
                              'space': 'space',
                              'pipeline': 'pipeline',
                              'experiment': 'experiment',
                              'member': 'space'}

            available_metanames = {'model': 'ModelMetaNames',
                                   'experiment': 'ExperimentMetaNames',
                                   'function': 'FunctionMetaNames',
                                   'pipeline': 'PipelineMetaNames'}

            actual_type = None

            for t in possible_types:
                if t in obj.__name__:
                    actual_type = possible_types[t]

            docs = f.__doc__

            if actual_type:
                docs = docs.replace(f'client.{actual_type}s.{f.__name__}', 'client.repository.' + obj.__name__)
                docs = docs.replace(f'client._{actual_type}s.{f.__name__}', 'client.repository.' + obj.__name__)

                if actual_type in available_metanames:
                    repository_meta_names = available_metanames[actual_type]
                    docs = docs.replace(f'_{actual_type}s.ConfigurationMetaNames', f'repository.{repository_meta_names}')
                    docs = docs.replace(f'{actual_type}s.ConfigurationMetaNames', f'repository.{repository_meta_names}')
                    docs = docs.replace('ConfigurationMetaNames', repository_meta_names)

                for k in mapping:
                    docs = docs.replace(k, mapping[k])
            obj.__doc__ = docs
        return obj
    return dec


def group_metrics(metrics):
    grouped_metrics = []

    if len(metrics) > 0:
        import collections
        grouped_metrics = collections.defaultdict(list)
        for d in metrics:
            k = d[u'phase']
            grouped_metrics[k].append(d)

    return grouped_metrics


class StatusLogger:
    def __init__(self, initial_state):
        self.last_state = initial_state
        print(initial_state, end='')

    def log_state(self, state):
        if state == self.last_state:
            print('.', end='')
        else:
            print('\n{}'.format(state), end='')
            self.last_state = state

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def get_file_from_cos(cos_credentials):
    import ibm_boto3
    from ibm_botocore.client import Config

    client_cos = ibm_boto3.client(service_name='s3',
        ibm_api_key_id=cos_credentials['IBM_API_KEY_ID'],
        ibm_auth_endpoint=cos_credentials['IBM_AUTH_ENDPOINT'],
        config=Config(signature_version='oauth'),
        endpoint_url=cos_credentials['ENDPOINT'])

    streaming_body = client_cos.get_object(Bucket=cos_credentials['BUCKET'], Key=cos_credentials['FILE'])['Body']
    training_definition_bytes = streaming_body.read()
    streaming_body.close()
    filename = cos_credentials['FILE']
    f = open(filename, 'wb')
    f.write(training_definition_bytes)
    f.close()

    return filename


def extract_model_from_repository(model_uid, client):
    """Download and extract archived model from wml repository.

    :param model_uid: UID of model
    :type model_uid: str
    :param client: client instance
    :type client: APIClient

    :return: extracted directory path
    :rtype: str
    """
    create_empty_directory(model_uid)
    current_dir = os.getcwd()

    os.chdir(model_uid)
    model_dir = os.getcwd()

    fname = 'downloaded_' + model_uid + '.tar.gz'
    client.repository.download(model_uid, filename=fname)

    if fname.endswith("tar.gz"):
        tar = tarfile.open(fname)
        tar.extractall()
        tar.close()
    else:
        raise WMLClientError('Invalid type. Expected tar.gz')

    os.chdir(current_dir)
    return model_dir


def extract_mlmodel_from_archive(archive_path, model_uid):
    """Extract archived model under model uid directory.

    :param model_uid: UID of model
    :type model_uid: str
    :param archive_path: path to archived model
    :type archive_path: str

    :return: extracted directory path
    :rtype: str
    """
    create_empty_directory(model_uid)
    current_dir = os.getcwd()

    os.rename(archive_path, os.path.join(model_uid, archive_path))

    os.chdir(model_uid)
    model_dir = os.getcwd()

    if archive_path.endswith("tar.gz"):
        tar = tarfile.open(archive_path)
        tar.extractall()
        tar.close()
    else:
        raise WMLClientError('Invalid type. Expected tar.gz')

    os.chdir(current_dir)
    return os.path.join(model_uid, 'model.mlmodel')


def get_model_filename(directory, model_extension):
    logger = logging.getLogger(__name__)
    model_filepath = None

    for file in os.listdir(directory):
        if file.endswith(model_extension):
            if model_filepath is None:
                model_filepath = os.path.join(directory, file)
            else:
                logger.warning('More than one file with extension \'{}\'.'.format(model_extension))

    if model_filepath is None:
        raise WMLClientError('No file with extension \'{}\'.'.format(model_extension))

    return model_filepath


def delete_directory(directory):
    if os.path.exists(directory):
        shutil.rmtree(directory)


def create_empty_directory(directory):
    delete_directory(directory)
    os.makedirs(directory)


def install_package(package):
    import importlib
    try:
        importlib.import_module(package)
    except ImportError:
        import pip
        pip.main(['install', package])


def is_ipython():
    # checks if the code is run in the notebook
    try:
        get_ipython
        return True
    except Exception:
        return False


def create_download_link(file_path, title="Download file."):
    # creates download link for binary files on notebook filesystem (Watson Studio)

    if is_ipython():
        from IPython.display import HTML
        import base64

        filename = os.path.basename(file_path)

        with open(file_path, 'rb') as file:
            b_model = file.read()
        b64 = base64.b64encode(b_model)
        payload = b64.decode()
        html = '<a download="{file_path}" href="data:binary;base64,{payload}" target="_blank">{title}</a>'
        html = html.format(payload=payload, title=title, file_path=filename)

        return HTML(html)


def convert_metadata_to_parameters(meta_data):
    parameters = []

    if meta_data is not None:
        if is_python_2():
            for key, value in meta_data.iteritems():
                parameters.append({'name': str(key), 'value': value})
        else:
            for key, value in meta_data.items():
                parameters.append({'name': str(key), 'value': value})

    return parameters


def is_of_python_basic_type(el):
    if type(el) in [int, float, bool, str]:
        return True
    elif type(el) in [list, tuple]:
        return all([is_of_python_basic_type(t) for t in el])
    elif type(el) is dict:
        if not all(type(k) == str for k in el.keys()):
            return False

        return is_of_python_basic_type(list(el.values()))
    else:
        return False


class NextResourceGenerator:
    """Generator class to produce next list of resources from REST API."""

    def __init__(self, wml_client: 'APIClient', url: str, href: str, params: dict = None, _all=False,
                 _filter_func=None) -> None:
        """
        :param wml_client: WML Client Instance
        :type wml_client: APIClient

        :param href: href to the resource
        :type href: str
        """
        self.wml_client = wml_client
        self.url = url
        self.next_href = href
        self.params = params
        self.all = _all
        self._filter_func = _filter_func

    def __iter__(self):
        while self.next_href is not None:
            if "http" not in self.next_href:
                self.next_href = f"{self.url}/{self.next_href}"
            response = requests.get(
                # url=f"{self.url}/{self.next_href}",
                url=self.next_href,
                headers=self.wml_client._get_headers(),
                params=self.params if self.params is not None else self.wml_client.params()
            )
            details_json = self.wml_client.training._handle_response(200, "Get next details", response)

            if self.all:
                self.next_href = details_json.get('next', {'href': None})['href']

            else:
                self.next_href = None

            if 'resources' in details_json:
                resources = details_json['resources']

            elif 'metadata' in details_json:
                resources = [details_json]

            else:
                resources = details_json.get('results', [])

            yield {'resources': self._filter_func(resources) if self._filter_func else resources}


class DisableWarningsLogger:
    """Class which disables logging warnings (for example for silent handling WMLClientErrors in try except).

    **Example**

    .. code-block:: python

        try:
            with DisableWarningsLogger():
                throw_wml_error()
        except WMLClientError:
            success = False

    """
    def __enter__(self):
        logging.disable(logging.WARNING)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)


def is_lib_installed(lib_name: str, minimum_version: Optional[str] = None, install: Optional[bool] = False) -> bool:
    """Check if provided library is installed on user environment. If not, tries to install it.

    :param lib_name: library name to check
    :type lib_name: str

    :param minimum_version: minimum version of library to check, default: None - check if library is installed in overall
    :type minimum_version: str, optional

    :param install: indicates to install missing or to low version library
    :type install: bool, optional

    :return: information if library is installed: `True` is library is installed, `False` otherwise
    :rtype: bool
    """
    if lib_name in sys.modules:
        installed = True

    elif importlib.util.find_spec(lib_name) is not None:
        installed = True

    else:
        installed = False

    if installed:
        installed_module_version = get_module_version(lib_name)

        if minimum_version is not None:
            if version.parse(installed_module_version) < version.parse(minimum_version):
                if install:
                    install_library(lib_name=lib_name, version=minimum_version, strict=False)

    else:
        if install:
            install_library(lib_name=lib_name, version=minimum_version, strict=False)
            installed = True

    return installed


def install_library(lib_name: str, version: Optional[str] = None, strict: Optional[bool] = False) -> None:
    """Try to install library.

    :param lib_name: library name to install
    :type lib_name: str

    :param version: version of the library to install
    :type version: str, optional

    :param strict: indicates if we want to install specific version or higher version if available
    :type strict: bool, optional
    """
    try:
        if version is not None:
            check_call([sys.executable, "-m", "pip", "install", f"{lib_name}{'==' if strict else '>='}{version}"])

        else:
            check_call([sys.executable, "-m", "pip", "install", lib_name])

    except Exception as e:
        raise CannotInstallLibrary(lib_name=lib_name, reason=str(e))


def get_module_version(lib_name: str) -> str:
    """Use only when you need to check package version by package name with pip."""
    from importlib.metadata import version

    return version(lib_name)


def prepare_interaction_props_for_cos(source_params: dict, file_name: str) -> dict:
    """If user specified properties for dataset as sheet_name, delimiter etc. we need to
    pass them as interaction properties for Flight Service.

    :param source_params: data source parameters describe data (eg. excel_sheet, encoding etc.)
    :type source_params: dict

    :param file_name: name of the file to download, should consist of file extension
    :type file_name: str

    :return: COS interaction properties for Flight Service
    :rtype: dict
    """
    interaction_properties = {}
    file_format = None

    encoding = source_params.get("encoding", None)

    if ".xls" in file_name or ".xlsx" in file_name:
        file_format = "excel"
        if source_params.get("excel_sheet"):
            interaction_properties["sheet_name"] = str(source_params.get("excel_sheet"))

    elif '.csv' in file_name:
        if source_params.get("quote_character"):
            interaction_properties["quote_character"] = str(source_params.get("quote_character"))
        if encoding is not None:
            interaction_properties["encoding"] = encoding

        input_file_separator = source_params.get("input_file_separator", ",")
        if input_file_separator != ",":
            file_format = "delimited"
            interaction_properties["field_delimiter"] = input_file_separator
        else:
            file_format = "csv"

    elif '.parquet' in file_name or '.prq' in file_name:
        file_format = 'parquet'

    if file_format is not None:
        interaction_properties["file_format"] = file_format

    return interaction_properties


def modify_details_for_script_and_shiny(details_from_get: dict) -> dict:
    """Add the href and id of and asset to the same position as it is returned from the POST method
    it allows the `get_id`/`get_href` method to work with details returned by GET method.

    :param details_from_get: details of script/shiny app acquired using GET method
    :type details_from_get: dict

    :return: details with 'guid' and 'href' key added to 'metadata'
    :rtype: dict
    """
    try:
        details_from_get[u'metadata'][u'href'] = details_from_get[u'href']
        details_from_get[u'metadata'][u'guid'] = details_from_get[u'metadata'][u'asset_id']
    except KeyError:
        pass

    return details_from_get


def is_lale_pipeline(pipeline):
    return type(pipeline).__module__ == 'lale.operators' and type(pipeline).__qualname__ == 'TrainedPipeline'


 
class NumpyTypeEncoder(json.JSONEncoder):
    """Extended json.JSONEncoder to encode correctly numpy types."""
    def default(self, obj):
        if isinstance(obj, numpy.integer):
            return int(obj)
        elif isinstance(obj, numpy.bool_):
            return bool(obj)
        elif isinstance(obj, numpy.floating):
            return None if numpy.isnan(obj) else float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        else:
            return super().default(obj)
        

def _check_if_import_from_watsonx_ai(parent_obj_module: str, source_module: str, obj: str):
    try:
        if 'ibm_watsonx_ai' in parent_obj_module:
            import_module = importlib.__import__(source_module.replace('ibm_watson_machine_learning', 'ibm_watsonx_ai'),  
                                                 fromlist=[obj])
            return getattr(import_module, obj)
        else:
            import_module = importlib.__import__(source_module, 
                                                 fromlist=[obj])
            return getattr(import_module, obj)
    except ModuleNotFoundError:
        import_module = importlib.__import__(source_module, 
                                             fromlist=[obj])
        return getattr(import_module, obj)


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function

import json

import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.metanames import HwSpecMetaNames
from ibm_watson_machine_learning.utils import HW_SPEC_DETAILS_TYPE
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.wml_resource import WMLResource

_DEFAULT_LIST_LENGTH = 50


class HwSpec(WMLResource):
    """Store and manage hardware specs."""

    ConfigurationMetaNames = HwSpecMetaNames()
    """MetaNames for Hardware Specification."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    def get_details(self, hw_spec_uid):
        """Get hardware specification details.

        :param hw_spec_uid: Unique id of the hardware spec
        :type hw_spec_uid: str

        :return: metadata of the hardware specifications
        :rtype: dict

        **Example**

        .. code-block:: python

            hw_spec_details = client.hardware_specifications.get_details(hw_spec_uid)

        """
        HwSpec._validate_type(hw_spec_uid, u'hw_spec_uid', str, True)
        if not self._ICP or self._client.ICP_PLATFORM_SPACES:
            response = requests.get(self._client.service_instance._href_definitions.get_hw_spec_href(hw_spec_uid), params=self._client._params(skip_space_project_chk=True),
                                    headers=self._client._get_headers())
        else:
            response = requests.get(self._client.service_instance._href_definitions.get_hw_spec_href(hw_spec_uid), params=self._client._params(),
                                      headers=self._client._get_headers())

        if response.status_code == 200:
            return self._get_required_element_from_response(self._handle_response(200, u'get hardware spec details', response))
        else:
            return self._handle_response(200, u'get hardware spec details', response)

    def store(self, meta_props):
        """Create a hardware specification.

        :type meta_props: dict

        :return: metadata of the created hardware specification
        :rtype: dict

        **Example**

        .. code-block:: python

            meta_props = {
                client.hardware_specifications.ConfigurationMetaNames.NAME: "custom hardware specification",
                client.hardware_specifications.ConfigurationMetaNames.DESCRIPTION: "Custom hardware specification creted with SDK",
                client.hardware_specifications.ConfigurationMetaNames.NODES:{"cpu":{"units":"2"},"mem":{"size":"128Gi"},"gpu":{"num_gpu":1}}
             }

            client.hardware_specifications.store(meta_props)

        """

        HwSpec._validate_type(meta_props, u'meta_props', dict, True)
        hw_spec_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client)

        hw_spec_meta_json = json.dumps(hw_spec_meta)
        href = self._client.service_instance._href_definitions.get_hw_specs_href()

        creation_response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(), data=hw_spec_meta_json)

        hw_spec_details = self._handle_response(201, u'creating hardware specification', creation_response)

        return hw_spec_details

    def list(self, name=None, return_as_df=True):
        """Print hardware specifications in a table format.

        :param name: Unique id of the hardware spec
        :type name: str, optional

        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed hardware specifications or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        **Example**

        .. code-block:: python

            client.hardware_specifications.list()

        """

        params = {}

        if name is not None:
            params.update({'name': name})

        # Todo provide api to return
        href = self._client.service_instance._href_definitions.get_hw_specs_href()

        response = requests.get(href, params, headers=self._client._get_headers())

        self._handle_response(200, u'list hw_specs', response)
        asset_details = self._handle_response(200, u'list assets', response)["resources"]
        hw_spec_values = [
            (m[u'metadata'][u'name'], m[u'metadata'][u'asset_id'], m[u'metadata'][u'description']) for
            m in asset_details]
        table = self._list(hw_spec_values, [u'NAME', u"ID", u'DESCRIPTION'], None, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table

    @staticmethod
    def get_id(hw_spec_details):
        """Get ID of hardware specifications asset.

        :param hw_spec_details: metadata of the hardware specifications
        :type hw_spec_details: dict

        :return: Unique Id of hardware specifications
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.hardware_specifications.get_id(hw_spec_details)

        """

        return HwSpec.get_uid(hw_spec_details)

    @staticmethod
    def get_uid(hw_spec_details):
        """Get UID of hardware specifications asset.

        *Deprecated:* Use ``get_id(hw_spec_details)`` instead.

        :param hw_spec_details: metadata of the hardware specifications
        :type hw_spec_details: dict

        :return: Unique Id of hardware specifications
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.hardware_specifications.get_uid(hw_spec_details)

        """
        HwSpec._validate_type(hw_spec_details, u'hw_spec_details', object, True)
        HwSpec._validate_type_of_details(hw_spec_details, HW_SPEC_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(hw_spec_details, u'hw_spec_details',
                                                           [u'metadata', u'asset_id'])


    @staticmethod
    def get_href(hw_spec_details):
        """Get url of hardware specifications.

        :param hw_spec_details: hardware specifications details
        :type hw_spec_details: dict

        :return: href of hardware specifications
        :rtype: str

        **Example**

        .. code-block:: python

            hw_spec_details = client.hw_spec.get_details(hw_spec_uid)
            hw_spec_href = client.hw_spec.get_href(hw_spec_details)

        """
        HwSpec._validate_type(hw_spec_details, u'hw_spec_details', object, True)
        HwSpec._validate_type_of_details(hw_spec_details, HW_SPEC_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(hw_spec_details, u'hw_spec_details', [u'metadata', u'href'])

    def get_id_by_name(self, hw_spec_name):
        """Get Unique Id of hardware specification for the given name.

        :param hw_spec_name: name of the hardware spec
        :type hw_spec_name: str

        :return: Unique Id of hardware specification
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.hardware_specifications.get_id_by_name(hw_spec_name)

        """

        return HwSpec.get_uid_by_name(self, hw_spec_name)

    def get_uid_by_name(self, hw_spec_name):
        """Get Unique Id of hardware specification for the given name.

        *Deprecated:* Use ``get_id_by_name(hw_spec_name)`` instead.

        :param hw_spec_name:  name of the hardware spec
        :type hw_spec_name: str

        :return: Unique Id of hardware specification
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.hardware_specifications.get_uid_by_name(hw_spec_name)

        """
        HwSpec._validate_type(hw_spec_name, u'hw_spec_name', str, True)
        parameters = self._client._params(skip_space_project_chk=True)
        parameters.update(name=hw_spec_name)

        response = requests.get(self._client.service_instance._href_definitions.get_hw_specs_href(),
                                params=parameters,
                                headers=self._client._get_headers())

        if response.status_code == 200:
            total_values = self._handle_response(200, u'list assets', response)["total_results"]
            if total_values != 0:
                hw_spec_details = self._handle_response(200, u'list assets', response)["resources"]
                return hw_spec_details[0][u'metadata'][u'asset_id']
            else:
                return "Not Found"

    def delete(self, hw_spec_id: str):
        """Delete a hardware specifications.

        :param hw_spec_id: Unique Id of hardware specification  which should be deleted
        :type hw_spec_id: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str
        """
        HwSpec._validate_type(hw_spec_id, u'hw_spec_id', str, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_hw_spec_href(hw_spec_id), params=self._client._params(),
                                headers=self._client._get_headers())

        return self._handle_response(204, u'delete hardware specification', response)


    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'hw_spec_response', dict)
        try:
            if self._client.default_space_id is not None:
                new_el = {'metadata': {
                                       'name': response_data['metadata']['name'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'href': response_data['metadata']['href'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at']
                                       #'updated_at': response_data['metadata']['updated_at']
                                       },
                          'entity': response_data['entity']

                          }
            elif self._client.default_project_id is not None:
                if self._client.WSD:
                    new_el = {'metadata': {
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'href': response_data['metadata']['href'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                           },
                              'entity': response_data['entity']

                              }
                else:
                    new_el = {'metadata': {
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'href': response_data['metadata']['href'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                       },
                             'entity': response_data['entity']

                            }
            else:
                # For system hardware spec
                new_el = {'metadata': {
                                       'name': response_data['metadata']['name'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'href': response_data['metadata']['href'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at']
                                       #'updated_at': response_data['metadata']['updated_at']
                                       },
                          'entity': response_data['entity']

                          }
            if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and 'href' in response_data['metadata']:
                href_without_host = response_data['metadata']['href'].split('.com')[-1]
                new_el[u'metadata'].update({'href': href_without_host})

            return new_el
        except Exception as e:
            raise WMLClientError("Failed to read Response from down-stream service: " + response_data.text)



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class BatchInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, artifact_version_href=None, name=None, input=None, output=None):
        """
        BatchInput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'artifact_version_href': 'str',
            'name': 'str',
            'input': 'InternalInputBatch',
            'output': 'InternalOutputBatch'
        }

        self.attribute_map = {
            'artifact_version_href': 'artifactVersionHref',
            'name': 'name',
            'input': 'input',
            'output': 'output'
        }

        self._artifact_version_href = artifact_version_href
        self._name = name
        self._input = input
        self._output = output

    @property
    def artifact_version_href(self):
        """
        Gets the artifact_version_href of this BatchInput.
        Href to the artifact version that is going to be deployed

        :return: The artifact_version_href of this BatchInput.
        :rtype: str
        """
        return self._artifact_version_href

    @artifact_version_href.setter
    def artifact_version_href(self, artifact_version_href):
        """
        Sets the artifact_version_href of this BatchInput.
        Href to the artifact version that is going to be deployed

        :param artifact_version_href: The artifact_version_href of this BatchInput.
        :type: str
        """

        self._artifact_version_href = artifact_version_href

    @property
    def name(self):
        """
        Gets the name of this BatchInput.
        User-friendly name for this Execution

        :return: The name of this BatchInput.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this BatchInput.
        User-friendly name for this Execution

        :param name: The name of this BatchInput.
        :type: str
        """

        self._name = name

    @property
    def input(self):
        """
        Gets the input of this BatchInput.


        :return: The input of this BatchInput.
        :rtype: InternalInputBatch
        """
        return self._input

    @input.setter
    def input(self, input):
        """
        Sets the input of this BatchInput.


        :param input: The input of this BatchInput.
        :type: InternalInputBatch
        """

        self._input = input

    @property
    def output(self):
        """
        Gets the output of this BatchInput.


        :return: The output of this BatchInput.
        :rtype: InternalOutputBatch
        """
        return self._output

    @output.setter
    def output(self, output):
        """
        Sets the output of this BatchInput.


        :param output: The output of this BatchInput.
        :type: InternalOutputBatch
        """

        self._output = output

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from enum import Enum

from ibm_watson_machine_learning.foundation_models.prompt_tuner import PromptTuner
from ibm_watson_machine_learning.foundation_models.utils.enums import PromptTuningTypes, PromptTuningInitMethods, \
    TuneExperimentTasks
from ibm_watson_machine_learning.foundation_models.utils.utils import get_model_specs_with_prompt_tuning_support, _check_model_state
from ibm_watson_machine_learning.experiment.base_experiment import BaseExperiment
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning import APIClient

from .tune_runs import TuneRuns


class TuneExperiment(BaseExperiment):
    """TuneExperiment class for tuning models with prompts.

    :param credentials: credentials to Watson Machine Learning instance
    :type credentials: dict

    :param project_id: ID of the Watson Studio project
    :type project_id: str, optional

    :param space_id: ID of the Watson Studio Space
    :type space_id: str, optional

    :param verify: user can pass as verify one of following:

        - the path to a CA_BUNDLE file
        - the path of directory with certificates of trusted CAs
        - `True` - default path to truststore will be taken
        - `False` - no verification will be made
    :type verify: bool or str, optional

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.experiment import TuneExperiment

        experiment = TuneExperiment(
            credentials={
                "apikey": "...",
                "iam_apikey_description": "...",
                "iam_apikey_name": "...",
                "iam_role_crn": "...",
                "iam_serviceid_crn": "...",
                "instance_id": "...",
                "url": "https://us-south.ml.cloud.ibm.com"
            },
            project_id="...",
            space_id="...")
    """

    def __init__(self,
                 credentials: dict,
                 project_id: str = None,
                 space_id: str = None,
                 verify=None) -> None:

        self.client = APIClient(credentials, verify=verify)
        if not self.client.CLOUD_PLATFORM_SPACES and self.client.CPD_version < 4.8:
            raise WMLClientError(error_msg="Operation is unsupported for this release.")

        if project_id:
            self.client.set.default_project(project_id)
        else:
            self.client.set.default_space(space_id)

        self.PromptTuningTypes = PromptTuningTypes
        self.PromptTuningInitMethods = PromptTuningInitMethods

        self.Tasks = self._get_tasks_enum()  # Note: Dynamically create enum with supported ENUM Tasks

        self.runs = TuneRuns(client=self.client)

    def runs(self, *, filter: str) -> 'TuneRuns':
        """Get the historical tuning runs but with name filter.

        :param filter: filter, user can choose which runs to fetch specifying tuning name
        :type filter: str

        **Examples**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(...)
            experiment.runs(filter='prompt tuning name').list()
        """
        return TuneRuns(client=self.client, filter=filter)

    def prompt_tuner(self,
                     name: str,  # Note: Rest API does not require name,
                     task_id: str,
                     description: str = None,
                     base_model: str = None,
                     accumulate_steps: int = None,
                     batch_size: int = None,
                     init_method: str = None,
                     init_text: str = None,
                     learning_rate: float = None,
                     max_input_tokens: int = None,
                     max_output_tokens: int = None,
                     num_epochs: int = None,
                     verbalizer: str = None,
                     tuning_type: str = None,
                     auto_update_model: bool = True,
                     group_by_name: bool = False) -> PromptTuner:
        """Initialize a PromptTuner module.

        :param name: name for the PromptTuner
        :type name: str

        :param task_id: task that is targeted for this model. Example: `experiment.Tasks.CLASSIFICATION`

            Possible values:

            - experiment.Tasks.CLASSIFICATION: 'classification' (default)
            - experiment.Tasks.QUESTION_ANSWERING: 'question_answering'
            - experiment.Tasks.SUMMARIZATION: 'summarization'
            - experiment.Tasks.RETRIEVAL_AUGMENTED_GENERATION: 'retrieval_augmented_generation'
            - experiment.Tasks.GENERATION: 'generation'
            - experiment.Tasks.CODE_GENERATION_AND_CONVERSION: 'code'
            - experiment.Tasks.EXTRACTION: 'extraction

        :type task_id: str

        :param description: description
        :type description: str, optional

        :param base_model: model id of the base model for this prompt tuning. Example: google/flan-t5-xl
        :type base_model: str, optional

        :param accumulate_steps: Number of steps to be used for gradient accumulation. Gradient accumulation 
            refers to a method of collecting gradient for configured number of steps instead of updating 
            the model variables at every step and then applying the update to model variables. 
            This can be used as a tool to overcome smaller batch size limitation. 
            Often also referred in conjunction with "effective batch size". Possible values: 1 ≤ value ≤ 128,
            default value: 16
        :type accumulate_steps: int, optional

        :param batch_size: The batch size is a number of samples processed before the model is updated. 
            Possible values: 1 ≤ value ≤ 16, default value: 16
        :type batch_size: int, optional

        :param init_method: `text` method requires `init_text` to be set. Allowable values: [`random`, `text`],
            default value: `random`
        :type init_method: str, optional

        :param init_text: initialization text to be used if `init_method` is set to `text` otherwise this will be ignored.
        :type init_text: str, optional

        :param learning_rate: learning rate to be used while tuning prompt vectors. Possible values: 0.01 ≤ value ≤ 0.5,
            default value: 0.3
        :type learning_rate: float, optional

        :param max_input_tokens: maximum length of input tokens being considered. Possible values: 1 ≤ value ≤ 256,
            default value: 256
        :type max_input_tokens: int, optional

        :param max_output_tokens: maximum length of output tokens being predicted. Possible values: 1 ≤ value ≤ 128
            default value: 128
        :type max_output_tokens: int, optional

        :param num_epochs: number of epochs to tune the prompt vectors, this affects the quality of the trained model.
            Possible values: 1 ≤ value ≤ 50, default value: 20
        :type num_epochs: int, optional

        :param verbalizer: verbalizer template to be used for formatting data at train and inference time. 
            This template may use brackets to indicate where fields from the data model must be rendered. 
            The default value is "{{input}}" which means use the raw text, default value: `Input: {{input}} Output:`
        :type verbalizer: str, optional

        :param tuning_type: type of Peft (Parameter-Efficient Fine-Tuning) config to build. 
            Allowable values: [`experiment.PromptTuningTypes.PT`], default value: `experiment.PromptTuningTypes.PT`
        :type tuning_type: str, optional

        :param auto_update_model: define if model should be automatically updated, default value: `True`
        :type auto_update_model: bool, optional

        :param group_by_name: define if tunings should be grouped by name, default value: `False`
        :type group_by_name: bool, optional

        **Examples**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import TuneExperiment

            experiment = TuneExperiment(...)

            prompt_tuner = experiment.prompt_tuner(
                name="prompt tuning name",
                task_id=experiment.Tasks.CLASSIFICATION,
                base_model='google/flan-t5-xl',
                accumulate_steps=32,
                batch_size=16,
                learning_rate=0.2,
                max_input_tokens=256,
                max_output_tokens=2,
                num_epochs=6,
                tuning_type=experiment.PromptTuningTypes.PT,
                verbalizer="Extract the satisfaction from the comment. Return simple '1' for satisfied customer or '0' for unsatisfied. Input: {{input}} Output: ",
                auto_update_model=True)
        """

        task_id, base_model = [enum_possible_field.value if isinstance(enum_possible_field, Enum) else enum_possible_field for enum_possible_field in (task_id, base_model)]

        prompt_tuning_supported_models = [model_spec['model_id'] for model_spec in
                                          get_model_specs_with_prompt_tuning_support(
                                              url=self.client.wml_credentials.get('url')).get('resources', [])]

        if base_model not in prompt_tuning_supported_models:
            raise WMLClientError(
                f"Base model '{base_model}' is not supported. Supported models: {prompt_tuning_supported_models}")
        
        # check if model is in constricted mode
        _check_model_state(self.client.wml_credentials.get('url'), base_model)

        prompt_tuner = PromptTuner(name=name,
                                   task_id=task_id,
                                   description=description,
                                   base_model=base_model,
                                   accumulate_steps=accumulate_steps,
                                   batch_size=batch_size,
                                   init_method=init_method,
                                   init_text=init_text,
                                   learning_rate=learning_rate,
                                   max_input_tokens=max_input_tokens,
                                   max_output_tokens=max_output_tokens,
                                   num_epochs=num_epochs,
                                   tuning_type=tuning_type,
                                   verbalizer=verbalizer,
                                   auto_update_model=auto_update_model,
                                   group_by_name=group_by_name)

        prompt_tuner._client = self.client

        return prompt_tuner

    def _get_tasks_enum(self):
        try:
            from ibm_watson_machine_learning.foundation_models.utils.utils import get_all_supported_tasks_dict
            return Enum(value='TuneExperimentTasks',
                        names=get_all_supported_tasks_dict(url=self.client.wml_credentials.get('url')))
        except Exception:
            return TuneExperimentTasks
